'' =================================================================================================
''
''   File....... hc-8+_vixen_to_dmx__161028.spin
''   Purpose.... Transmit Vixen files via DMX on HC-8+
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (C) 2016 Jon McPhalen
''   E-mail..... 
''   Started.... 
''   Updated.... 28 OCT 2016
''
'' =================================================================================================


con

  ' Important: This program is designed to be used on Rev F and later boards
  '            -- do not use with Rev E or earlier.

  PCB_REV = "F"


con { timing }

  _clkmode = xtal1 + pll16x
  _xinfreq = 5_000_000                                          ' use 5MHz crystal

  CLK_FREQ = _xinfreq * (_clkmode >> 6)                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us
                                                                 
                                                                 
con { io pins }                                                  
                                                                 
  RX1     = 31  { I }                                           ' serial / programming
  TX1     = 30  { O }                                            
                                                                 
  SDA     = 29  { I/O }                                         ' i2c / eeprom
  SCL     = 28  { I/O }                                          
                                                                 
  XP0     = 27  { I/O }                                         ' extra port 0 - not on Rev E or lower
  XP1     = 24  { I/O }                                         ' extra port 1 - not on Rev E or lower 
  XP2     = 19  { I/O }                                         ' extra port 2 - not on Rev E or lower    
  XP3     = 16  { I/O }                                         ' extra port 3 - not on Rev E or lower 
                                                                 
  SIO_RX  = 26  { I/O }                                         ' TTL serial io
  SIO_TX  = 25  { I/O }                                         ' TTL serial io (not enabled)  
                                                                 
  RX2     = 23  { I }                                           ' RS-485 port
  RX2_EN  = 22  { O }                                                
  TX2_EN  = 21  { O }                                                
  TX2     = 20  { O }                                                
  
  G_LED   = 18  { O }                                           ' bi-color LED (Rev F assignments)
  R_LED   = 17  { O }                                              
                                                                 
  OUT7    = 15  { O }                                           ' outputs
  OUT6    = 14  { O }                                                
  OUT5    = 13  { O }                                                
  OUT4    = 12  { O }                                                
  OUT3    = 11  { O }                                                
  OUT2    = 10  { O }                                                
  OUT1    =  9  { O }                                                
  OUT0    =  8  { O }                                                
                                                                 
  OPT_BR  =  7  { I/O }                                         ' options switches
  OPT_A1  =  6  { I/O }                                                 
  OPT_A0  =  5  { I/O }                                                 
  OPT_SM  =  4  { I/O }                                                 
                                                                 
  DMX_A8  =  3  { I }                                           ' high-bit of DMX address
                                                                 
  DO_165  =  2  { I }                                           ' control lines for 74x165
  CLK_165 =  1  { O }                                                 
  LD_165  =  0  { O }                                                 


  ' for EFX-TEK uSD adapter module (#99002)
  ' -- must open all option switches when installed

  SD_CS   =  OPT_BR
  SD_DI   =  OPT_A1
  SD_CLK  =  OPT_A0
  SD_DO   =  OPT_SM
                                               
                                                                 
con                                                              
                                                                 
  CHANNELS = 64
                                                     
                                                                 
  #(-1), ON, OFF                                                ' digital control
                                                                 
  #0, NO, YES                                                    
                                                                 
  #0, BLK, RED, GRN, YEL                                        ' for r/g led
                                                                 
  #0, NOCON, RX, TX, RXTX                                       ' for rs-485
                                                                 
  #0, LSBFIRST, MSBFIRST                                         
                                                                 
                                                                 
obj                              
                                 
' main                                                          ' * master Spin cog                   
  time       : "jm_time"                                        '   timing and delays  
  io         : "jm_io_basic"                                    '   essential io
  dmx        : "jm_dmx_tx"                                      ' * DMX output
  sd         : "fsrw"                                           ' * SD card IO
' background                                                    ' * background inputs scanning
                                                                 
' * uses cog when loaded
                                 
                                 
var

  { ---- Vixen File Header ---- }
  { Do not change order or size }

  word  filechannels                                            ' channels in Vixen file
  word  eventms                                                 ' event timing; milliseconds
  word  frames                                                  ' frames (events) in Vixen file

  { --------------------------- }

  byte  framebuf[CHANNELS]                                      ' buffer for raw frame data


dat

  ShowFile      byte    "vixen00.bin", 0                        ' base show file name

                                          
pub main | show                                                 
                                                                 
  setup                                                         ' setup io and objects

  ' Look for uSD card on HC-8+
  ' -- make sure all OPTION switches are set to OFF
  ' -- waits for uSD insertion
  ' -- flash at 1Hz if no card found

  repeat
    if (mount_sd == false)
      set_rgled(RED)
      time.pause(500)
      set_rgled(OFF)
      time.pause(500)

    else
      set_rgled(OFF)
      quit

  ' main program loop
      
  repeat
    show := ttl_pin                                             ' look for single input
    if ((show => 0) and (show =< 7))
      play_show(show)                                           ' if pressed, run it

    time.pause(1000)                                            ' inter-program delay


pub play_show(show) | frametix, t, ch, tix

'' Play show from uSD card

  ShowFile[6] := "0" + show                                     ' update filename
                                                                 
  ifnot (open_file(@ShowFile, "r"))                             ' open file
    close_file                                                   
    return                                                       
                                                                 
  ifnot (read_buf(@filechannels, 6))                            ' get header
    close_file                                                   
    return                                                       
                                                                 
  if (filechannels > CHANNELS)                                  ' okay to play?
    close_file                                                   
    return                                                       
                                                                 
  bytefill(@framebuf, 0, CHANNELS)                              ' clear old buffer
                                                                 
  frametix := MS_001 * eventms                                  ' calculate ticks per frame
                                                                 
  set_rgled(GRN)
  io.high(OUT0 + show)                                          ' indicate show playing                                                
                                                                 
  t := cnt                                                      ' sync frame timer
  repeat frames                                                 ' loop through frames
    if (read_buf(@framebuf, filechannels))                      ' read new frame
      bytemove(dmx.address+1, @framebuf, filechannels)          ' copy to DMX output
      waitcnt(t += frametix)                                    ' hold for frame time
      io.toggle(G_LED)

    else
      quit

  close_file
  
  set_rgled(OFF)
  outa[OUT7..OUT0] := %00000000                                 ' indicator off
  
                                                           
pub setup
                                                                 
'' Configure HC-8+ IO and support objects                        
                                                                 
  time.start                                                    ' setup for timing/delays
                                                                 
  io.start(0, 0)                                                ' clear pins (this cog)
                                                                 
  set_rs485(NOCON)                                              ' RS-485 not connected

  dmx.startx(TX2, TX2_EN, 0, CHANNELS, 5_0)                     ' refresh DMX every 5ms (100 ch)                                                                      

  bcog := cognew(background, @bcstack) + 1                      ' start background cog *


con

  { ----------------- }
  {  F I L E   I / 0  }
  { ----------------- }


var

  long  mounted


pub mount_sd | check

'' Attempt to mount SD card
'' -- returns true if SD card successfully mounted

  check := \sd.mount_explicit(SD_DO, SD_CLK, SD_DI, SD_CS)

  if (check => 0)
    mounted := true
  else
    mounted := false

  return mounted


pub dismount_sd

'' Unmounts SD card

  \sd.unmount
  mounted := false


pub open_file(p_str, mode) | check

'' Attempt to open file
'' -- p_str is pointer to filename (z-string)
'' -- mode is single character, (e.g., "r" for read)
''    * must be lowercase
'' -- sd card must be mounted before open

  if (mounted)
    check := \sd.popen(p_str, mode)                             ' attempt to open
    if (check => 0)
      if (mode == "r")
        \sd.seek(0)                                             ' force sector read
      return true
    else
      return false
  else
    return false


pub close_file

'' Closes open file

  \sd.pclose


pub read_buf(p_buf, n) | check

'' Reads n bytes from file to buffer at bpntr
'' -- returns true if good read

  check := \sd.pread(p_buf, n)
  if (check == n)
    return true
  else
    return false


con

  { --------------------------------- }
  {  S U P P O R T   R O U T I N E S  } 
  { --------------------------------- }


var

  long  ttlpins                                                 ' updated by background cog
  long  dmxaddr                                                  
                                                                 
                                                                 
pub ttl_pin                                                      
                                                                 
'' Reads TTL states for 1-of-8 input                             
'' -- returns -1 (no input)                                      
'' -- returns 0 to 7 for input on IN0 to IN7                     
''    * only one input allowed                                   
                                                                 
  if (bit_count(ttlpins) == 1)                                  ' only one button pressed?
    return bit_pos(ttlpins, 0)                                  '  yes, return button #
  else                                                           
    return -1                                                   '  none or bad input
                                                                 

pub ttl_inputs 

'' Return TTL inputs status
'' -- pins are scanned by background cog

  return ttlpins


pub dmx_address 

'' Return DMX address switch setting
'' -- switches are scanned by background cog
'' -- Note: For Rev C and lower boards, do not use DMX 256 switch -- leave it OFF

  return dmxaddr


pub option_inputs

'' Returns state of option switch bits (%0000..%1111, all off to all on)

  dira[OPT_BR..OPT_SM] := %0000                                 ' force to input mode
                                                                  
  return !ina[OPT_BR..OPT_SM] & $0F                             ' return state (1 = on)
                                                                  

pub xport_bits

'' Returns state of xport bits (%0000..%1111, all off to all on)
'' -- does not set dira bits to 0

  return (ina[XP3] << 3) | (ina[XP2] << 2) | (ina[XP1] << 1) | ina[XP0]   

                                                         
pub set_rgled(color)                                              
                                                                  
'' Sets R/G LED                                                   
'' -- direct control only                                         
'' -- red/green/yellow + blinking requires support cog   
                                                                  
  case color                                                    ' set color to state                  
    OFF : outa[R_LED..G_LED] := %00                                                                   
    RED : outa[R_LED..G_LED] := %10                                                                   
    GRN : outa[R_LED..G_LED] := %01                                                                   
    YEL : outa[R_LED..G_LED] := %00                             ' yellow requires active driver       
                                                                 
  dira[R_LED..G_LED] := %11                                     ' make outputs


pub set_rs485(state)                                              
                                                                  
'' Sets RS-485 RX and TX enable inputs                            
                                                                  
  case state                                                      
    NOCON : outa[RX2_EN..TX2_EN] := %10                           
    RX    : outa[RX2_EN..TX2_EN] := %00                           
    TX    : outa[RX2_EN..TX2_EN] := %11                           
    RXTX  : outa[RX2_EN..TX2_EN] := %01                           
                                                                  
  dira[RX2_EN..TX2_EN] := %11                                     
                                                                  
                                                                  
pub bit_count(value) | bc                                         
                                                                  
'' Returns # of bits set (1) in value                             
                                                                  
  bc := 0                                                       ' clear count
  repeat 32                                                     ' test all bits
    bc += (value & %1)                                          ' add bit value to count
    value >>= 1                                                 ' next bit
                                                                 
  return bc                                                      
                                                                 
                                                                 
pub bit_pos(value, mode)                                         
                                                                 
'' Returns position of 1st bit set (0..31)                       
'' -- mode 0 to scan from lsb, mode 1 to scan from msb           
'' -- -1 = no bits set                                           
                                                                 
  if (value == 0)                                               ' if no bits                                   
    return -1                                                   '  return -1
  else                                                           
    if (mode == LSBFIRST)                                       ' check from LSB
      value ><= 32                                              '  flip for >|
      return (32 - >|value)                                      
    else                                                          
      return (>|value - 1)                                        
                                                                  
                                                                  
pub bit_val(value, pos)                                           
                                                            
'' Returns bit (0..1) from pos (0..31) in value                   
                                                                  
  if ((pos => 0) and (pos =< 31))                                 
    return (value >> pos) & 1                                     
  else                                                            
    return 0


con

  { ----------------------------- }
  {  B A C K G R O U N D   C O G  }
  {  - global milliseconds        }
  {  - TTL/DMX inputs scanning    }
  { ----------------------------- }
  

var  

  long  bcog                                                    ' background cog #
  long  bcstack[64]                                             ' stack for background cog
                                                                 
  long  millis                                                  ' global milliseconds register
                                                                 
                                                                 
pri background | t                                              ' launch with cognew()
                                                                 
  io.low(R_LED)                                                 ' setup R/G LED pins
  io.low(G_LED)                                                  
                                                                 
  io.high(LD_165)                                               ' setup x165 io pins
  io.low(CLK_165)                                                
  io.input(DO_165)                                               
                                                                 
  millis := 0                                                    
                                                                 
  t := cnt                                                      ' sync loop timer
  repeat                                                         
    waitcnt(t += MS_001)                                        ' run loop every millisecond
    ++millis                                                     
    scan_ttl_ins                                                ' re-scan TTL & DMX inputs
                                                                 
                                                               
pri scan_ttl_ins | tempin                                         
                                                                 
'' Scan TTL and DMX address inputs                               
                                                                 
  outa[LD_165] := 0                                             ' blip Shift/Load line    
  outa[LD_165] := 1                                              

  ' unrolled for best speed

  tempin := (tempin << 1) | ina[DO_165]                         ' bit15
  outa[CLK_165] := 1                                            ' blip clock  
  outa[CLK_165] := 0                                                        
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0                   
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0                   
  tempin := (tempin << 1) | ina[DO_165]
  outa[CLK_165] := 1                   
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
  tempin := (tempin << 1) | ina[DO_165] 
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                                       
  tempin := (tempin << 1) | ina[DO_165]                         ' bit0
  outa[CLK_165] := 1                    
  outa[CLK_165] := 0                    
                                                         
  ttlpins := tempin.byte[0]                                     ' update global vars
  dmxaddr := tempin.byte[1] | (ina[DMX_A8] << 8) 


dat { license }                                                  

{{

  Copyright (C) 2016 Jon McPhalen

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}      